/************************************************************************
    MeOS - Orienteering Software
    Copyright (C) 2009-2011 Melin Software HB
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Melin Software HB - software@melin.nu - www.melin.nu
    Stigbergsvgen 7, SE-75242 UPPSALA, Sweden
    
************************************************************************/

#include "stdafx.h"
#include "oListInfo.h"
#include "oEvent.h"
#include "gdioutput.h"
#include "meos_util.h"
#include <cassert>
#include <cmath>
#include "Localizer.h"
#include "metalist.h"
#include <algorithm>
#include "meosexception.h"
#include "gdifonts.h"

MetaList::MetaList() {
  data.resize(4);
  hasResults_ = false;
  initSymbols();
}

bool MetaList::isBreak(int x) const {
  return isspace(x) || x == '.' || x == ',' || 
          x == '-'  || x == ':' || x == ';' || x == '('
          || x == ')' || x=='/';
}

string MetaList::encode(const string &input_) const {
  string out;
  string input = lang.tl(input_);
  out.reserve(input.length() + 5);

  for (size_t k = 0; k<input.length(); k++) {
    int c = input[k];
    int p = k > 0 ? input[k-1] : ' ';
    int n = k+1 < input.length() ? input[k+1] : ' ';

    if (c == '%') {
      out.push_back('%');
      out.push_back('%');
    }
    else if (c == 'X' &&  isBreak(n) && isBreak(p) ) {
      out.push_back('%');
      out.push_back('s');
    }
    else
      out.push_back(c);
  }
  return out;
}

MetaListPost &MetaList::add(ListIndex ix, const MetaListPost &post) {
  if (data[ix].empty())
    addRow(ix);
  //  data[ix].resize(1);
  
  vector<MetaListPost> &d = data[ix].back();
  d.push_back(post);
  return d.back();
}

void MetaList::addRow(int ix) {
  data[ix].push_back(vector<MetaListPost>());
}

#ifndef MEOSDB

void MetaList::interpret(oEvent *oe, const oListParam &par, 
                         int lineHeight, oListInfo &li) const {
  const MetaList &mList = *this;
  Position pos;
  bool large = par.useLargeSize;

  gdiFonts normal, header, small;
  double s;

  if (large) {
    s = 1.8;
    normal = fontLarge;
    header = boldLarge;
    small = normalText;
    lineHeight = int(lineHeight *1.6);
  }
  else {
    s = 1.0;
    normal = normalText;
    header = boldText;
    small = italicSmall;
  }

  map<EPostType, string> labelMap;
  map<string, string> stringLabelMap;
  
  set<EPostType> skip;
  li.calcResults = false;
  li.calcTotalResults = false;
  li.rogainingResults = false;

  for (int i = 0; i<4; i++) {
    const vector< vector<MetaListPost> > &lines = mList.data[i];
    for (size_t j = 0; j<lines.size(); j++) {
      const vector<MetaListPost> &cline = lines[j];
      for (size_t k = 0; k<cline.size(); k++) {
        const MetaListPost &mp = cline[k];

        // Automatically determine what needs to be calculated
        if (mp.type == lTeamPlace || mp.type == lRunnerPlace) {
          if (!li.calcResults)
            oe->calculateResults(oEvent::RTClassResult);
          li.calcResults = true;
        }
        else if (mp.type == lRunnerTotalPlace) {
          if (!li.calcTotalResults)
            oe->calculateResults(oEvent::RTTotalResult);

          li.calcTotalResults = true;
        }
        else if (mp.type == lRunnerRogainingPoint) {
          //if (!li.rogainingResults)
          //  oe->calculateRogainingResults();
          li.rogainingResults = true;
        }
        string label = "P" + itos(i*1000 + j*100 + k);
        
        if (mp.type == lRunnerBib || mp.type == lTeamBib) {
          if (!oe->hasBib(mp.type == lRunnerBib, mp.type == lTeamBib))
            skip.insert(mp.type);
        }

        if (mp.type == lTeamStart || mp.type == lRunnerStart ||
             mp.type == lClassStartTime) {
          bool hasIndStart = false;
          
          for (set<int>::const_iterator it = par.selection.begin(); it != par.selection.end(); ++it) {
            pClass pc = oe->getClass(*it);
            if (pc) {
              int first, last;
              pc->getStartRange(par.legNumber, first, last);
              if (last != first) {
                hasIndStart = true;
                break;
              }
            }
          }
          if (mp.type == lClassStartTime) {
            if (hasIndStart)
              skip.insert(mp.type);
          }
          else {
            if (!hasIndStart)
              skip.insert(mp.type);
          }
        }

        int width = 0;
        if (skip.count(mp.type) == 0) {
          width = li.getMaxCharWidth(oe, mp.type, encode(mp.text), large, mp.blockWidth);
        }

        if (mp.alignType == lNone) {
          pos.add(label, width);
          pos.indent(mp.minimalIndent);
        }
        else {
          if (mp.alignType == lAlignNext)
            pos.alignNext(label, width, mp.alignBlock);
          else if (mp.alignType == lString) {
            if (stringLabelMap.count(mp.alignWithText) == 0) {
              throw meosException("Don't know how to align with X#" + mp.alignWithText);
            }
            pos.update(stringLabelMap[mp.alignWithText], label, width, mp.alignBlock);
          }
          else {
            if (labelMap.count(mp.alignType) == 0) {
              throw meosException("Don't know how to align with X#" + typeToSymbol[mp.alignType]);
            }

            pos.update(labelMap[mp.alignType], label, width, mp.alignBlock);
          }

          pos.indent(mp.minimalIndent);
        }
        labelMap[mp.type] = label;
        if (!mp.text.empty())
          stringLabelMap[mp.text] = label;

      }
      pos.newRow();
    }
    pos.newRow();
  }

  bool c = true;
  while (c) {
    c = pos.postAdjust();
  }

  int dx = 0, next_dx = 0;

  if (large == false && par.pageBreak == false) {
    for (size_t j = 0; j<mList.getHead().size(); j++) {
      const vector<MetaListPost> &cline = mList.getHead()[j];
      for (size_t k = 0; k<cline.size(); k++) {
        const MetaListPost &mp = cline[k];
        if (skip.count(mp.type) == 1)
          continue;
          
        string label = "P" + itos(0*1000 + j*100 + k);
        string text = encode(cline[k].text);
        gdiFonts font = normalText;
        if (mp.type == lCmpName) {
          font = boldLarge;
          next_dx = lineHeight*2;
          text = MakeDash(par.getCustomTitle(text));
        }
        
        li.addHead(oPrintPost(cline[k].type, text, font, 
                              pos.get(label), dx, cline[k].leg));
      }
      dx += next_dx;
      next_dx = lineHeight;
    }
  }

  for (size_t j = 0; j<mList.getSubHead().size(); j++) {
    const vector<MetaListPost> &cline = mList.getSubHead()[j];
    for (size_t k = 0; k<cline.size(); k++) {
      const MetaListPost &mp = cline[k];
      if (skip.count(mp.type) == 1)
        continue;
        
      string label = "P" + itos(1*1000 + j*100 + k);
      li.addSubHead(oPrintPost(cline[k].type, encode(cline[k].text), header, 
                                pos.get(label, s), lineHeight*(j+1), cline[k].leg));
    }
  }

  int sub_dy = mList.getSubList().size() > 0 ? lineHeight/2 : 0;

  for (size_t j = 0; j<mList.getList().size(); j++) {
    const vector<MetaListPost> &cline = mList.getList()[j];
    for (size_t k = 0; k<cline.size(); k++) {
      const MetaListPost &mp = cline[k];
      if (skip.count(mp.type) == 1)
        continue;
        
      string label = "P" + itos(2*1000 + j*100 + k);
      li.addListPost(oPrintPost(cline[k].type, encode(cline[k].text), normal, 
                                pos.get(label, s), lineHeight*j + sub_dy, cline[k].leg));
    }
  }

  for (size_t j = 0; j<mList.getSubList().size(); j++) {
    const vector<MetaListPost> &cline = mList.getSubList()[j];
    for (size_t k = 0; k<cline.size(); k++) {
      const MetaListPost &mp = cline[k];
      if (skip.count(mp.type) == 1)
        continue;
        
      string label = "P" + itos(3*1000 + j*100 + k);
      li.addSubListPost(oPrintPost(cline[k].type, encode(cline[k].text), small, 
                                pos.get(label, s), lineHeight*j, cline[k].leg));
    }
  }

  li.listType = listType;
  li.listSubType = listSubType;
  li.sortOrder = sortOrder;
  for (set<EFilterList>::const_iterator it = filter.begin(); 
                                      it != filter.end(); ++it) {
    li.setFilter(*it);
  }
}

#endif

void Position::indent(int ind) {
  int end = pos.size() - 1;
  if (end < 1)
    return;
  if (pos[end-1] < ind) {
    int dx = ind - pos[end-1];
    pos[end-1] += dx;
    pos[end] += dx;
  }
}

void Position::newRow() {
  if (!pos.empty())
    pos.pop_back();
  pos.push_back(0);
}

void Position::alignNext(const string &newname, int width, bool alignBlock) {
  if (pos.empty())
    return;
  int p = pos.empty() ? 0 : pos.back();
  const int backAlign = 20;
  const int fwdAlign = max(width/2, 40);

  if (p==0)
    p = backAlign;

  int next = 0, prev = 0;
  int next_p = 100000, prev_p = -100000;

  int last = pos.size()-1;
  for (int k = pos.size()-2; k >= 0; k--) {
    last = k;
    if (pos[k+1] < pos[k])
      break;
  }

  for (int k = 0; k <= last; k++) {
    if ( pos[k] >= p && pos[k] < next_p ) {
      next = k;
      next_p = pos[k];
    }

    if ( pos[k] < p && pos[k] > prev_p ) {
      prev = k;
      prev_p = pos[k];
    }
  }

  if ( p - prev_p < backAlign) {
    int delta = p - prev_p;
    for (size_t k = 0; k + 1 < pos.size(); k++) {
      if (pos[k] >= prev_p)
        pos[k] += delta;
    }
    update(prev, newname, width, alignBlock);
  }
  else {
    if (next > 0 && (next_p - p) < fwdAlign)
      update(next, newname, width, alignBlock);
    else
      add(newname, width);
  }
}

void Position::update(const string &oldname, const string &newname, int width, bool alignBlock) {
  if (pmap.count(oldname) == 0)
    throw std::exception("Invalid position");

  int ix = pmap[oldname];
  
  update(ix, newname, width, alignBlock);
}

void Position::update(int ix, const string &newname, int width, bool alignBlock) {

  int last = pos.size()-1;
  int xlimit = pos[ix];
  if (xlimit < pos[last]) {
    int delta = pos[last] - xlimit;
    
    // Find last entry to update (higher row)
    int lastud = last;
    while (lastud>1 && pos[lastud] >= pos[lastud-1])
      lastud--;

    for (int k = 0; k<lastud; k++) {
      if (pos[k] >= xlimit)
        pos[k] += delta;
    }
  }
  else
    pos[last] = pos[ix];

  int ow = pos[ix+1] - pos[ix];
  int nw = width;
  if (alignBlock && ow>0) {
    nw = max(ow, nw);
    if (nw > ow) {
      int delta = nw - ow;
      for (size_t k = 0; k<pos.size(); k++) {
        if (pos[k] > pos[ix])
          pos[k] += delta;
      }
    }
  }
  //pos.push_back(pos[ix]);
  add(newname, nw);  
}
  
void Position::add(const string &name, int width) {
  
  if (pos.empty()) {
    pos.push_back(0);
    pmap[name] = 0;
    pos.push_back(width);
  }
  else {
    pmap[name] = pos.size() - 1;
    pos.push_back(width + pos.back());
  }
}

int Position::get(const string &name) {
  return pos[pmap[name]];
}

int Position::get(const string &name, double scale) {
  return int(pos[pmap[name]] * scale);
}

struct PosOrderIndex {
  int pos;
  int index;
  int width;
  int row;
  bool operator<(const PosOrderIndex &x) const {return pos < x.pos;}
};

bool Position::postAdjust() {
  
  vector<PosOrderIndex> x;
  int row = 0;
  for (size_t k = 0; k < pos.size(); k++) {
    PosOrderIndex poi;
    poi.pos = pos[k];
    poi.index = k;
    poi.row = row;
    if (k + 1 == pos.size() || pos[k+1] == 0) {
      poi.width = 100000;
      row++;
    }
    else
      poi.width = pos[k+1] - pos[k];

    x.push_back(poi);
  }

  sort(x.begin(), x.end());

  pair<int, int> smallDiff(100000, -1);
  for (size_t k = 0; k<x.size(); k++) {
    if (k>0 && x[k-1].pos == x[k].pos || x[k].pos == 0)
      continue;
    int diff = 0;
    for (size_t j = k + 1; j < x.size(); j++) {
      if (x[j].pos > x[k].pos) {
        if (x[j].row == x[k].row)
          break;
        else {
          diff = x[j].pos - x[k].pos;

          bool skipRow = false;
          for (size_t i = 0; i<x.size(); i++) {
            if (x[i].pos == x[k].pos && x[i].row == x[j].row) {
              skipRow = true;
              break;
            }
          }

          if (skipRow || diff > x[k].width / 2)
            break;

          if (diff < smallDiff.first) {
            smallDiff.first = diff;
            smallDiff.second = k;
          }
        }
      }
    }
  }
  
  if (smallDiff.second != -1) {
    int minK = smallDiff.second;
    int diff = smallDiff.first;
    int basePos = x[minK].pos;
    set<int> keepRow;
    for (size_t k = 0; k<x.size(); k++) {
      if (x[k].pos == basePos + diff) {
        keepRow.insert(x[k].row);
      }
    }
    bool changed = false;
    assert(keepRow.count(x[minK].row) == 0);
    for (size_t k = 0; k<x.size(); k++) {
      if (x[k].pos >= basePos && keepRow.count(x[k].row) == 0) {
        pos[x[k].index] += diff;
        changed = true;
      }
    }

    return changed;
  }
  return false;
}

void MetaList::save(const string &file) const {
  xmlparser xml;
  xml.openOutput(file.c_str(), true);
  save(xml);
  xml.closeOut();
}

void MetaList::save(xmlparser &xml) const {
  xml.startTag("MeOSListDefinition");
//  xml.write("Title", defaultTitle);
  xml.write("ListName", listName);
  xml.write("Tag", tag);
  xml.write("SortOrder", orderToSymbol[sortOrder]);
  xml.write("ListType", baseTypeToSymbol[listType]);
  xml.write("SubListType", baseTypeToSymbol[listSubType]);

  //xml.writeBool("CalculateResults", calcResults);
  //xml.writeBool("Rogaining", rogainingResults);

  for (set<EFilterList>::const_iterator it = filter.begin(); it != filter.end(); ++it)
    xml.write("Filter", "name", filterToSymbol[*it]);

  serialize(xml, "Head", getHead());
  serialize(xml, "SubHead", getSubHead());
  serialize(xml, "List", getList());
  serialize(xml, "SubList", getSubList());
  
  xml.endTag();
}

void MetaList::load(const string &file) {
  xmlparser xml;
  xml.read(file.c_str());
  filter.clear();
  xmlobject xDef = xml.getObject("MeOSListDefinition");
  load(xDef);
}

void MetaList::load(const xmlobject &xDef) {
  xDef.getObjectString("ListName", listName);
  xDef.getObjectString("Tag", tag);
  
  string tmp;
  xDef.getObjectString("SortOrder", tmp);
  if (symbolToOrder.count(tmp) == 0) {
    string err = "Invalid sort order X#" + tmp;
    throw std::exception(err.c_str());
  }
  sortOrder = symbolToOrder[tmp];

  xDef.getObjectString("ListType", tmp);
  if (symbolToBaseType.count(tmp) == 0) {
    string err = "Invalid type X#" + tmp;
    throw std::exception(err.c_str());
  }

  listType = symbolToBaseType[tmp];
  if (symbolToBaseType.count(tmp) == 0) {
    string err = "Invalid type X#" + tmp;
    throw std::exception(err.c_str());
  }

  xDef.getObjectString("SubListType", tmp);

  listSubType = symbolToBaseType[tmp];
  //calcResults = xDef.getObjectBool("CalculateResults");
  //rogainingResults = xDef.getObjectBool("Rogaining");

  xmlobject xHead = xDef.getObject("Head");
  xmlobject xSubHead = xDef.getObject("SubHead");
  xmlobject xList = xDef.getObject("List");
  xmlobject xSubList = xDef.getObject("SubList");

  deserialize(xHead, data[MLHead]);
  deserialize(xSubHead, data[MLSubHead]);
  deserialize(xList, data[MLList]);
  deserialize(xSubList, data[MLSubList]);

  // Check if result list
  for (int i = 0; i<4; i++) {
    const vector< vector<MetaListPost> > &lines = data[i];
    for (size_t j = 0; j<lines.size(); j++) {
      const vector<MetaListPost> &cline = lines[j];
      for (size_t k = 0; k<cline.size(); k++) {
        const MetaListPost &mp = cline[k];

        if (mp.type == lTeamPlace || mp.type == lRunnerPlace) {
          hasResults_ = true;
        }
        else if (mp.type == lRunnerTotalPlace) {
          hasResults_ = true;
        }
        else if (mp.type == lRunnerRogainingPoint) {
          hasResults_ = true;
        }
      }
    }
  }


  xmlList f;
  xDef.getObjects("Filter", f);

  for (size_t k = 0; k<f.size(); k++) {
    string attrib = f[k].getAttrib("name").get();
    if (symbolToFilter.count(attrib) == 0) {
      string err = "Invalid filter X#" + attrib; 
      throw std::exception(err.c_str());
    }
    addFilter(symbolToFilter[attrib]);
  }
}

void MetaList::serialize(xmlparser &xml, const string &tagp, 
                         const vector< vector<MetaListPost> > &lp) const {
  xml.startTag(tagp.c_str());

  for (size_t k = 0; k<lp.size(); k++) {
    xml.startTag("Line");
    for (size_t j = 0; j<lp[k].size(); j++) {
      lp[k][j].serialize(xml);
    }
    xml.endTag();
  }

  xml.endTag();
}

void MetaList::deserialize(const xmlobject &xml, vector< vector<MetaListPost> > &lp) {
  xmlList xLines;
  xml.getObjects(xLines);

  for (size_t k = 0; k<xLines.size(); k++) {
    lp.push_back(vector<MetaListPost>());
    xmlList xBlocks;
    xLines[k].getObjects(xBlocks);
    for (size_t j = 0; j<xBlocks.size(); j++) {
      lp[k].push_back(MetaListPost(lNone));
      lp[k].back().deserialize(xBlocks[j]);
    }
  }
}

void MetaListPost::serialize(xmlparser &xml) const {
  xml.startTag("Block", "Type", MetaList::typeToSymbol[type]);
  xml.write("Text", text);
  xml.write("Leg", leg);
  if (alignType == lString)
    xml.write("Align", "BlockAlign", alignBlock, alignWithText);
  else
    xml.write("Align", "BlockAlign", alignBlock, MetaList::typeToSymbol[alignType]);
  xml.write("BlockWidth", blockWidth);
  xml.write("IndentMin", minimalIndent);
  xml.endTag();
}

void MetaListPost::deserialize(const xmlobject &xml) {
  string tp = xml.getAttrib("Type").get();
  if (MetaList::symbolToType.count(tp) == 0) {
    string err = "Invalid type X#" + tp; 
    throw std::exception(err.c_str());
  }

  type = MetaList::symbolToType[tp];
  xml.getObjectString("Text", text);
  leg = xml.getObjectInt("Leg");
  xmlobject xAlignBlock = xml.getObject("Align");
  alignBlock = xAlignBlock && xAlignBlock.getObjectBool("BlockAlign"); 
  blockWidth = xml.getObjectInt("BlockWidth");
  minimalIndent = xml.getObjectInt("IndentMin");
  string at;
  xml.getObjectString("Align", at);

  if (!at.empty()) {
    if (MetaList::symbolToType.count(at) == 0) {
      //string err = "Invalid align type X#" + at; 
      //throw std::exception(err.c_str());
      alignType = lString;
      alignWithText = at;
    }
    else alignType = MetaList::symbolToType[at];
  }
}


map<EPostType, string> MetaList::typeToSymbol;
map<string, EPostType> MetaList::symbolToType;
map<oListInfo::EBaseType, string> MetaList::baseTypeToSymbol;
map<string, oListInfo::EBaseType> MetaList::symbolToBaseType;
map<SortOrder, string> MetaList::orderToSymbol;
map<string, SortOrder> MetaList::symbolToOrder;
map<EFilterList, string> MetaList::filterToSymbol;
map<string, EFilterList> MetaList::symbolToFilter;
    

void MetaList::initSymbols() {
  if (typeToSymbol.empty()) {
    typeToSymbol[lAlignNext] = "AlignNext";
    typeToSymbol[lNone] = "None";
    typeToSymbol[lString] = "String";
    typeToSymbol[lCmpName] = "CmpName";
    typeToSymbol[lCmpDate] = "CmpDate";
    typeToSymbol[lCurrentTime] = "CurrentTime";
    typeToSymbol[lClubName] = "ClubName";
    typeToSymbol[lClassName] = "ClassName";
    typeToSymbol[lClassStartName] = "ClassStartName";
    typeToSymbol[lClassStartTime] = "ClassStartTime";
    typeToSymbol[lClassLength] = "ClassLength";
    typeToSymbol[lClassResultFraction] = "ClassResultFraction";
    typeToSymbol[lRunnerName] = "RunnerName";
    typeToSymbol[lRunnerGivenName] = "RunnerGivenName";
    typeToSymbol[lRunnerFamilyName] = "RunnerFamilyName";
    typeToSymbol[lRunnerCompleteName] = "RunnerCompleteName";
    typeToSymbol[lPatrolNameNames] = "PatrolNameNames";
    typeToSymbol[lRunnerFinish] = "RunnerFinish";
    typeToSymbol[lRunnerTime] = "RunnerTime";
    typeToSymbol[lRunnerTimeStatus] = "RunnerTimeStatus";
    typeToSymbol[lRunnerTempTimeStatus] = "RunnerTempTimeStatus";
    typeToSymbol[lRunnerTempTimeAfter] = "RunnerTempTimeAfter";
    typeToSymbol[lRunnerTimeAfter] = "RunnerTimeAfter";
    typeToSymbol[lRunnerMissedTime] = "RunnerTimeLost";
    typeToSymbol[lRunnerPlace] = "RunnerPlace";
    typeToSymbol[lRunnerStart] = "RunnerStart";
    typeToSymbol[lRunnerClub] = "RunnerClub";
    typeToSymbol[lRunnerCard] = "RunnerCard";
    typeToSymbol[lRunnerBib] = "RunnerBib";
    typeToSymbol[lRunnerStartNo] = "RunnerStartNo";
    typeToSymbol[lRunnerRank] = "RunnerRank";
    typeToSymbol[lRunnerCourse] = "RunnerCourse";
    typeToSymbol[lRunnerRogainingPoint] = "RunnerRogainingPoint";
    typeToSymbol[lRunnerUMMasterPoint] = "RunnerUMMasterPoint";
    typeToSymbol[lRunnerTimePlaceFixed] = "RunnerTimePlaceFixed";
    typeToSymbol[lTeamName] = "TeamName";
    typeToSymbol[lTeamStart] = "TeamStart";
    typeToSymbol[lTeamTimeStatus] = "TeamTimeStatus";
    typeToSymbol[lTeamTimeAfter] = "TeamTimeAfter";
    typeToSymbol[lTeamPlace] = "TeamPlace";
    typeToSymbol[lTeamLegTimeStatus] = "TeamLegTimeStatus";
    typeToSymbol[lTeamLegTimeAfter] = "TeamLegTimeAfter";
    typeToSymbol[lTeamRogainingPoint] = "TeamRogainingPoint";
    typeToSymbol[lTeamTime] = "TeamTime";
    typeToSymbol[lTeamStatus] = "TeamStatus";
    typeToSymbol[lTeamClub] = "TeamClub";
    typeToSymbol[lTeamRunner] = "TeamRunner";
    typeToSymbol[lTeamRunnerCard] = "TeamRunnerCard";
    typeToSymbol[lTeamBib] = "TeamBib";
    typeToSymbol[lTeamStartNo] = "TeamStartNo";
    typeToSymbol[lPunchNamedTime] = "PunchNamedTime";
    typeToSymbol[lPunchTime] = "PunchTime";
    typeToSymbol[lRogainingPunch] = "RogainingPunch";
    typeToSymbol[lTotalCounter] = "TotalCounter";
    typeToSymbol[lSubCounter] = "SubCounter";
    typeToSymbol[lSubSubCounter] = "SubSubCounter";

    typeToSymbol[lRunnerTotalTime] = "RunnerTotalTime";
    typeToSymbol[lRunnerTotalTimeStatus] = "RunnerTotalTimeStatus";
    typeToSymbol[lRunnerTotalPlace] = "RunnerTotalPlace";
    typeToSymbol[lRunnerTotalTimeAfter] = "RunnerTotalTimeAfter";
    typeToSymbol[lRunnerTimeAfterDiff] = "RunnerTimeAfterDiff";
    typeToSymbol[lRunnerPlaceDiff] = "RunnerPlaceDiff";

    for (map<EPostType, string>::iterator it = typeToSymbol.begin();
      it != typeToSymbol.end(); ++it) {
      symbolToType[it->second] = it->first;
    }

    if (typeToSymbol.size() != lLastItem)
      throw std::exception("Bad symbol setup");

    if (symbolToType.size() != lLastItem)
      throw std::exception("Bad symbol setup");

    baseTypeToSymbol[oListInfo::EBaseTypeRunner] = "Runner";
    baseTypeToSymbol[oListInfo::EBaseTypeTeam] = "Team";
    baseTypeToSymbol[oListInfo::EBaseTypeClub] = "Club";
    baseTypeToSymbol[oListInfo::EBaseTypePunches] = "Punches";
    baseTypeToSymbol[oListInfo::EBaseTypeNone] = "None";

    for (map<oListInfo::EBaseType, string>::iterator it = baseTypeToSymbol.begin();
      it != baseTypeToSymbol.end(); ++it) {
      symbolToBaseType[it->second] = it->first;
    }

    if (baseTypeToSymbol.size() != oListInfo::EBasedTypeLast_)
      throw std::exception("Bad symbol setup");

    if (symbolToBaseType.size() != oListInfo::EBasedTypeLast_)
      throw std::exception("Bad symbol setup");

    orderToSymbol[ClassStartTime] = "ClassStartTime";
    orderToSymbol[ClassStartTimeClub] = "ClassStartTimeClub";
    orderToSymbol[ClassResult] = "ClassResult";
    orderToSymbol[SortByName] = "Name";
    orderToSymbol[SortByFinishTime] = "FinishTime";
    orderToSymbol[SortByClassFinishTime] = "ClassFinishTime";
    orderToSymbol[SortByStartTime] = "StartTime";
    orderToSymbol[SortByPoints] = "Points";
    orderToSymbol[ClassTotalResult] = "ClassTotalResult";
    orderToSymbol[CourseResult] = "CourseResult";
   
    for (map<SortOrder, string>::iterator it = orderToSymbol.begin();
      it != orderToSymbol.end(); ++it) {
      symbolToOrder[it->second] = it->first;
    }

    if (orderToSymbol.size() != SortEnumLastItem)
      throw std::exception("Bad symbol setup");

    if (symbolToOrder.size() != SortEnumLastItem)
      throw std::exception("Bad symbol setup");

    filterToSymbol[EFilterHasResult] = "FilterResult";
    filterToSymbol[EFilterRentCard] = "FilterRentCard";
    filterToSymbol[EFilterHasCard] = "FilterHasCard";
    filterToSymbol[EFilterExcludeDNS] = "FilterStarted";
    filterToSymbol[EFilterVacant] = "FilterNotVacant";
    filterToSymbol[EFilterHasNoCard] = "FilterNoCard";

    for (map<EFilterList, string>::iterator it = filterToSymbol.begin();
      it != filterToSymbol.end(); ++it) {
      symbolToFilter[it->second] = it->first;
    }

    if (filterToSymbol.size() != _EFilterMax)
      throw std::exception("Bad symbol setup");

    if (symbolToFilter.size() != _EFilterMax)
      throw std::exception("Bad symbol setup");

  }
}

MetaListContainer::MetaListContainer() {}

MetaListContainer::~MetaListContainer() {}


const MetaList &MetaListContainer::getList(int index) const {
  return data[index].second;
}

MetaList &MetaListContainer::getList(int index) {
  return data[index].second;
}

MetaList &MetaListContainer::addExternal(const MetaList &ml) {
  data.push_back(make_pair(ExternalList, ml));
  return data.back().second;
}

void MetaListContainer::clearExternal() {
  globalIndex.clear();
  while(!data.empty() && data.back().first == ExternalList)
    data.pop_back();
}

void MetaListContainer::save(MetaListType type, xmlparser &xml) const {
  for (size_t k = 0; k<data.size(); k++) {
    if (data[k].first == type)
      data[k].second.save(xml);
  }
}

void MetaListContainer::load(MetaListType type, const xmlobject &xDef) {
  xmlList xList;
  xDef.getObjects("MeOSListDefinition", xList);
  
  if (xList.empty() && strcmp(xDef.getName(), "MeOSListDefinition") == 0)
    xList.push_back(xDef);
  
  for (size_t k = 0; k<xList.size(); k++) {
    data.push_back(make_pair(type, MetaList()));
    data.back().second.load(xList[k]);
  }
}

#ifndef MEOSDB
void MetaListContainer::setupListInfo(int firstIndex, 
                                      map<EStdListType, oListInfo> &listMap, 
                                      bool resultsOnly) const {
  globalIndex.clear();
  for (size_t k = 0; k<data.size(); k++) {
    const MetaList &ml = data[k].second;
    if (!resultsOnly || ml.hasResults()) {
      EStdListType listIx = EStdListType(k + firstIndex);
      oListInfo &li = listMap[listIx];

      if (data[k].first == InternalList) {
        const string &tag = data[k].second.getTag();
        if (!tag.empty())
          tagIndex[tag] = listIx;
      }


      globalIndex[listIx] = k;
      li.Name = lang.tl(ml.getListName());
      li.listType = ml.getListType();
      li.supportClasses = true;
      li.supportLegs = true;
      li.supportExtra = false;
    }
  }
}

bool MetaListContainer::interpret(oEvent *oe, const oListParam &par, 
                                  int lineHeight, oListInfo &li) const {
                                    
  map<EStdListType, int>::const_iterator it = globalIndex.find(par.listCode);
  if (it != globalIndex.end()) {
    data[it->second].second.interpret(oe, par, lineHeight, li);
    return true;
  }
  return false;
}

EStdListType MetaListContainer::getType(const std::string &tag) const {
  map<string, EStdListType>::iterator it = tagIndex.find(tag);

  if (it == tagIndex.end())
    throw meosException("Could not load list 'X'.#" + tag);

  return it->second;
}

#endif
