#pragma once

/************************************************************************
    MeOS - Orienteering Software
    Copyright (C) 2009-2011 Melin Software HB
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Melin Software HB - software@melin.nu - www.melin.nu
    Stigbergsvgen 7, SE-75242 UPPSALA, Sweden
    
************************************************************************/

class oListInfo;
enum EPostType;
#include "oListInfo.h"

class Position 
{
  map<string, int> pmap;
  vector<int> pos;
  void update(int ix, const string &newname, int width, bool alignBlock);

public:
  bool postAdjust();

  void add(const string &name, int width);
  void update(const string &oldname, const string &newname, int width, bool alignBlock);
  void alignNext(const string &newname, int width, bool alignBlock);

  void newRow();
  
  int get(const string &name);
  int get(const string &name, double scale);

  void indent(int ind);
};

class MetaListPost {
private:
  void serialize(xmlparser &xml) const;
  void deserialize(const xmlobject &xml);

  EPostType type;
  string text;
  string alignWithText;
  EPostType alignType;
  int leg;
  int minimalIndent;
  bool alignBlock; // True if the next item should also be align (table style block)
  int blockWidth;

public:

  MetaListPost(EPostType type_, EPostType align_ = lNone, int leg_ = 0) : type(type_), alignType(align_), leg(leg_), minimalIndent(0), alignBlock(true), blockWidth(0) {}
  
  MetaListPost &setBlock(int width) {blockWidth = width; return *this;}
  MetaListPost &setText(const string &text_) {text = text_; return *this;}
  MetaListPost &align(EPostType align_, bool alignBlock_ = true) {alignType = align_; alignBlock = alignBlock_; return *this;}
  MetaListPost &align(bool alignBlock_ = true) {return align(lAlignNext, alignBlock_);}
  MetaListPost &indent(int ind) {minimalIndent = ind; return *this;}
  
 friend class MetaList;
};

class MetaList {
private:
  vector< vector< vector<MetaListPost> > > data;
  string listName;
  string tag;
  bool hasResults_;

  oListInfo::EBaseType listType;
  oListInfo::EBaseType listSubType;
  SortOrder sortOrder;

  set<EFilterList> filter;

  const vector< vector<MetaListPost> > &getList() const {return data[MLList];}
  const vector< vector<MetaListPost> > &getSubList() const {return data[MLSubList];}
  const vector< vector<MetaListPost> > &getHead() const {return data[MLHead];}
  const vector< vector<MetaListPost> > &getSubHead() const {return data[MLSubHead];}

  enum ListIndex {MLHead = 0, MLSubHead = 1, MLList = 2, MLSubList=3};
  MetaListPost &add(ListIndex ix, const MetaListPost &post);
  void addRow(int ix);
  string encode(const string &input) const;
  bool isBreak(int x) const;

  static map<EPostType, string> typeToSymbol;
  static map<string, EPostType> symbolToType;
  
  static map<oListInfo::EBaseType, string> baseTypeToSymbol;
  static map<string, oListInfo::EBaseType> symbolToBaseType;
  
  static map<SortOrder, string> orderToSymbol;
  static map<string, SortOrder> symbolToOrder;
  
  static map<EFilterList, string> filterToSymbol;
  static map<string, EFilterList> symbolToFilter;

  void initSymbols();

  void serialize(xmlparser &xml, const string &tag, 
                 const vector< vector<MetaListPost> > &lp) const;
  void deserialize(const xmlobject &xml, vector< vector<MetaListPost> > &lp);

public:
  MetaList();
  virtual ~MetaList() {}

  const string &getListName() const {return listName;}
  oListInfo::EBaseType getListType() const {return listType;}
  bool hasResults() const {return hasResults_;}
  const string &getTag() const {return tag;}

//  MetaList &hasResults(bool r) {calcResults = r; return *this;}
//  MetaList &isRogaining(bool r) {rogainingResults = r; return *this;}
  MetaList &setListType(oListInfo::EBaseType t) {listType = t; return *this;}
  MetaList &setSubListType(oListInfo::EBaseType t) {listSubType = t; return *this;}
  MetaList &setSortOrder(SortOrder so) {sortOrder = so; return *this;}

  MetaList &addFilter(EFilterList f) {filter.insert(f); return *this;}

  void save(const string &file) const;
  void load(const string &file);

  void save(xmlparser &xml) const;
  void load(const xmlobject &xDef);

  void interpret(oEvent *oe, const oListParam &par, 
                 int lineHeight, oListInfo &li) const;

  MetaList &setListName(const string &title) {listName = title; return *this;}

  MetaListPost &addToHead(const MetaListPost &post) {return add(MLHead, post);}
  MetaListPost &addToSubHead(const MetaListPost &post) {return add(MLSubHead, post).setBlock(10);}
  MetaListPost &addToList(const MetaListPost &post) {return add(MLList, post);}
  MetaListPost &addToSubList(const MetaListPost &post) {return add(MLSubList, post);}

  void newListRow() {addRow(MLList);}
  void newSubListRow() {addRow(MLSubList);}
  void newHead() {addRow(MLHead);}
  void newSubHead() {addRow(MLSubHead);}

  friend class MetaListPost;
};

class MetaListContainer {
public:
  enum MetaListType {InternalList, ExternalList};
private:
  vector< pair<MetaListType, MetaList> > data;
  mutable map<EStdListType, int> globalIndex;
  mutable map<string, EStdListType> tagIndex;
  
public:

  MetaListContainer();
  virtual ~MetaListContainer();
 
  EStdListType getType(const string &tag) const;

  const MetaList &getList(int index) const;
  MetaList &getList(int index);

  MetaList &addExternal(const MetaList &ml);
  void clearExternal();

  void save(MetaListType type, xmlparser &xml) const;
  void load(MetaListType type, const xmlobject &xDef);
  
  void setupListInfo(int firstIndex, map<EStdListType, oListInfo> &listMap, bool resultsOnly) const;

  bool interpret(oEvent *oe, const oListParam &par, 
                 int lineHeight, oListInfo &li) const;
};
